from __future__ import annotations
import numpy
import typing
from . import tf

__all__ = [
    "Arrow",
    "FastCrossing",
    "FlatBush",
    "KdQuiver",
    "KdTree",
    "LineSegment",
    "PolylineRuler",
    "Quiver",
    "densify_polyline",
    "douglas_simplify",
    "douglas_simplify_indexes",
    "douglas_simplify_mask",
    "intersect_segments",
    "point_in_polygon",
    "polyline_in_polygon",
    "snap_onto_2d",
    "tf",
]

class Arrow:
    @staticmethod
    def _angle(
        vec: numpy.ndarray[numpy.float64[3, 1]],
        *,
        ref: numpy.ndarray[numpy.float64[3, 1]],
    ) -> float:
        """
        Calculate angle between two vectors
        """
    @staticmethod
    @typing.overload
    def _heading(heading: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Convert heading to unit vector
        """
    @staticmethod
    @typing.overload
    def _heading(east: float, north: float) -> float:
        """
        Convert east and north components to heading
        """
    @staticmethod
    def _unit_vector(
        vector: numpy.ndarray[numpy.float64[3, 1]], with_eps: bool = True
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Normalize a vector to unit length
        """
    def Frenet(self) -> numpy.ndarray[numpy.float64[3, 3]]:
        """
        Get the Frenet frame of the Arrow
        """
    def __copy__(self, arg0: dict) -> Arrow:
        """
        Create a copy of the Arrow
        """
    @typing.overload
    def __init__(self) -> None:
        """
        Default constructor for Arrow
        """
    @typing.overload
    def __init__(self, position: numpy.ndarray[numpy.float64[3, 1]]) -> None:
        """
        Constructor for Arrow with position
        """
    @typing.overload
    def __init__(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        direction: numpy.ndarray[numpy.float64[3, 1]],
    ) -> None:
        """
        Constructor for Arrow with position and direction
        """
    def __repr__(self) -> str: ...
    def copy(self) -> Arrow:
        """
        Create a copy of the Arrow
        """
    @typing.overload
    def direction(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the direction of the Arrow
        """
    @typing.overload
    def direction(self, arg0: numpy.ndarray[numpy.float64[3, 1]]) -> Arrow:
        """
        Set the direction of the Arrow
        """
    def forward(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the forward direction of the Arrow
        """
    def has_index(self, check_range: bool = True) -> bool:
        """
        Check if the Arrow has a valid index
        """
    @typing.overload
    def heading(self) -> float:
        """
        Get the heading of the Arrow
        """
    @typing.overload
    def heading(self, new_value: float) -> Arrow:
        """
        Set the heading of the Arrow
        """
    @typing.overload
    def label(self) -> numpy.ndarray[numpy.int32[2, 1]]:
        """
        Get the label of the Arrow
        """
    @typing.overload
    def label(self, new_value: numpy.ndarray[numpy.int32[2, 1]]) -> Arrow:
        """
        Set the label of the Arrow
        """
    @typing.overload
    def label(
        self,
        polyline_index: int,
        segment_index: int,
        *,
        t: float | None = None,
        range: float | None = None,
    ) -> Arrow:
        """
        Set the label of the Arrow with polyline and segment indices
        """
    def leftward(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the leftward direction of the Arrow
        """
    @typing.overload
    def polyline_index(self) -> int:
        """
        Get the polyline index of the Arrow
        """
    @typing.overload
    def polyline_index(self, new_value: int) -> Arrow:
        """
        Set the polyline index of the Arrow
        """
    @typing.overload
    def position(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the position of the Arrow
        """
    @typing.overload
    def position(self, new_value: numpy.ndarray[numpy.float64[3, 1]]) -> Arrow:
        """
        Set the position of the Arrow
        """
    @typing.overload
    def range(self) -> float:
        """
        Get the range of the Arrow
        """
    @typing.overload
    def range(self, new_value: float) -> Arrow:
        """
        Set the range of the Arrow
        """
    def reset_index(self) -> None:
        """
        Reset the index of the Arrow
        """
    @typing.overload
    def segment_index(self) -> int:
        """
        Get the segment index of the Arrow
        """
    @typing.overload
    def segment_index(self, new_value: int) -> Arrow:
        """
        Set the segment index of the Arrow
        """
    @typing.overload
    def t(self) -> float:
        """
        Get the t parameter of the Arrow
        """
    @typing.overload
    def t(self, new_value: float) -> Arrow:
        """
        Set the t parameter of the Arrow
        """
    def upward(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the upward direction of the Arrow
        """

class FastCrossing:
    def __init__(self, *, is_wgs84: bool = False) -> None:
        """
        Initialize FastCrossing object.

        :param is_wgs84: Whether coordinates are in WGS84 format, defaults to false
        """
    @typing.overload
    def add_polyline(
        self, polyline: numpy.ndarray[numpy.float64[m, 3]], *, index: int = -1
    ) -> int:
        """
        Add polyline to the tree.

        :param polyline: The polyline to add
        :param index: Custom polyline index, defaults to -1
        :return: The index of the added polyline
        """
    @typing.overload
    def add_polyline(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        *,
        index: int = -1,
    ) -> int:
        """
        Add polyline to the tree (alternative format).

        :param polyline: The polyline to add
        :param index: Custom polyline index, defaults to -1
        :return: The index of the added polyline
        """
    def arrow(
        self, *, polyline_index: int, point_index: int
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], numpy.ndarray[numpy.float64[3, 1]]]:
        """
        Get an arrow (position and direction) at a specific point on a polyline.

        :param polyline_index: Index of the polyline
        :param point_index: Index of the point within the polyline
        :return: Arrow (position and direction)
        """
    def bush(self, autobuild: bool = True) -> ...:
        """
        Export the internal FlatBush index.

        :param autobuild: Whether to automatically build the index if not already built, defaults to true
        :return: FlatBush index
        """
    @typing.overload
    def coordinates(
        self, polyline_index: int, segment_index: int, ratio: float
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get coordinates at a specific position on a polyline.

        :param polyline_index: Index of the polyline
        :param segment_index: Index of the segment within the polyline
        :param ratio: Ratio along the segment (0 to 1)
        :return: Coordinates at the specified position
        """
    @typing.overload
    def coordinates(
        self, index: numpy.ndarray[numpy.int32[2, 1]], ratio: float
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get coordinates at a specific position on a polyline (alternative format).

        :param index: Combined index of polyline and segment
        :param ratio: Ratio along the segment (0 to 1)
        :return: Coordinates at the specified position
        """
    @typing.overload
    def coordinates(
        self,
        intersection: tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ],
        second: bool = True,
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get coordinates of an intersection.

        :param intersection: The intersection object
        :param second: Whether to use the second polyline of the intersection, defaults to true
        :return: Coordinates of the intersection
        """
    def finish(self) -> None:
        """
        Finalize indexing after adding all polylines
        """
    @typing.overload
    def intersections(
        self,
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get all segment intersections in the tree
        """
    @typing.overload
    def intersections(
        self, *, z_offset_range: tuple[float, float], self_intersection: int = 2
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get segment intersections in the tree, filtered by conditions.

        :param z_offset_range: Z-offset range for filtering
        :param self_intersection: Self-intersection parameter, defaults to 2
        """
    @typing.overload
    def intersections(
        self,
        start: numpy.ndarray[numpy.float64[2, 1]],
        to: numpy.ndarray[numpy.float64[2, 1]],
        *,
        dedup: bool = True,
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get crossing intersections with [start, to] segment.

        :param from: Start point of the segment
        :param to: End point of the segment
        :param dedup: Whether to remove duplicates, defaults to true
        :return: Sorted intersections by t ratio
        """
    @typing.overload
    def intersections(
        self, polyline: numpy.ndarray[numpy.float64[m, 3]], *, dedup: bool = True
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get crossing intersections with a polyline.

        :param polyline: The polyline to check intersections with
        :param dedup: Whether to remove duplicates, defaults to true
        :return: Sorted intersections by t ratio
        """
    @typing.overload
    def intersections(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        *,
        dedup: bool = True,
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get crossing intersections with a polyline (alternative format).

        :param polyline: The polyline to check intersections with
        :param dedup: Whether to remove duplicates, defaults to true
        :return: Sorted intersections by t ratio
        """
    @typing.overload
    def intersections(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 3]],
        *,
        z_min: float,
        z_max: float,
        dedup: bool = True,
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get crossing intersections with a polyline, filtered by Z range.

        :param polyline: The polyline to check intersections with
        :param z_min: Minimum Z value for filtering
        :param z_max: Maximum Z value for filtering
        :param dedup: Whether to remove duplicates, defaults to true
        :return: Sorted intersections by t ratio
        """
    @typing.overload
    def intersections(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        *,
        z_min: float,
        z_max: float,
        dedup: bool = True,
    ) -> list[
        tuple[
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.float64[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
            numpy.ndarray[numpy.int32[2, 1]],
        ]
    ]:
        """
        Get crossing intersections with a polyline, filtered by Z range (alternative format).

        :param polyline: The polyline to check intersections with
        :param z_min: Minimum Z value for filtering
        :param z_max: Maximum Z value for filtering
        :param dedup: Whether to remove duplicates, defaults to true
        :return: Sorted intersections by t ratio
        """
    def is_wgs84(self) -> bool:
        """
        Check if the coordinates are in WGS84 format.

        :return: True if coordinates are in WGS84 format, False otherwise
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[2, 1]], float]:
        """
        Find the nearest point to a given position.

        :param position: The query position
        :param return_squared_l2: Whether to return squared L2 distance, defaults to false
        :return: Nearest point information
        """
    @typing.overload
    def nearest(
        self,
        index: numpy.ndarray[numpy.int32[2, 1]],
        *,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[2, 1]], float]:
        """
        Find the nearest point to a given index.

        :param index: The query index
        :param return_squared_l2: Whether to return squared L2 distance, defaults to false
        :return: Nearest point information
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        k: int | None = None,
        radius: float | None = None,
        sort: bool = True,
        return_squared_l2: bool = False,
        filter: tuple[numpy.ndarray[numpy.float64[3, 1]], ...] | None = None,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 2]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Find k nearest points to a given position with optional filtering.

        :param position: The query position
        :param k: Number of nearest neighbors to find (optional)
        :param radius: Search radius (optional)
        :param sort: Whether to sort the results, defaults to true
        :param return_squared_l2: Whether to return squared L2 distance, defaults to false
        :param filter: Optional filter parameters
        :return: Nearest points information
        """
    def num_polylines(self) -> int:
        """
        Get the number of polylines in the FastCrossing object.

        :return: Number of polylines
        """
    @typing.overload
    def point_index(self, index: int) -> numpy.ndarray[numpy.int32[2, 1]]:
        """
        Get point index for a given index.

        :param index: The index to query
        :return: The point index
        """
    @typing.overload
    def point_index(
        self, indexes: numpy.ndarray[numpy.int32[m, 1]]
    ) -> list[numpy.ndarray[numpy.int32[2, 1]]]:
        """
        Get point indexes for given indexes.

        :param indexes: The indexes to query
        :return: The point indexes
        """
    def polyline_ruler(self, index: int) -> ...:
        """
        Get a specific polyline ruler.

        :param index: Index of the polyline
        :return: Polyline ruler for the specified index
        """
    def polyline_rulers(self) -> dict[int, ...]:
        """
        Get all polyline rulers.

        :return: Dictionary of polyline rulers
        """
    def quiver(self) -> ...:
        """
        Export the internal Quiver object.

        :return: Quiver object
        """
    @typing.overload
    def segment_index(self, index: int) -> numpy.ndarray[numpy.int32[2, 1]]:
        """
        Get segment index for a given index.

        :param index: The index to query
        :return: The segment index
        """
    @typing.overload
    def segment_index(
        self, indexes: numpy.ndarray[numpy.int32[m, 1]]
    ) -> list[numpy.ndarray[numpy.int32[2, 1]]]:
        """
        Get segment indexes for given indexes.

        :param indexes: The indexes to query
        :return: The segment indexes
        """
    @typing.overload
    def within(
        self,
        *,
        min: numpy.ndarray[numpy.float64[2, 1]],
        max: numpy.ndarray[numpy.float64[2, 1]],
        segment_wise: bool = True,
        sort: bool = True,
    ) -> list[numpy.ndarray[numpy.int32[2, 1]]]:
        """
        Find polylines within a bounding box.

        :param min: Minimum corner of the bounding box
        :param max: Maximum corner of the bounding box
        :param segment_wise: Whether to return segment-wise results, defaults to true
        :param sort: Whether to sort the results, defaults to true
        :return: Polylines within the bounding box
        """
    @typing.overload
    def within(
        self,
        *,
        polygon: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        segment_wise: bool = True,
        sort: bool = True,
    ) -> list[numpy.ndarray[numpy.int32[2, 1]]]:
        """
        Find polylines within a polygon.

        :param polygon: The polygon to check against
        :param segment_wise: Whether to return segment-wise results, defaults to true
        :param sort: Whether to sort the results, defaults to true
        :return: Polylines within the polygon
        """
    @typing.overload
    def within(
        self,
        *,
        center: numpy.ndarray[numpy.float64[2, 1]],
        width: float,
        height: float,
        heading: float = 0.0,
        segment_wise: bool = True,
        sort: bool = True,
    ) -> list[numpy.ndarray[numpy.int32[2, 1]]]:
        """
        Find polylines within a rotated rectangle.

        :param center: Center of the rectangle
        :param width: Width of the rectangle
        :param height: Height of the rectangle
        :param heading: Heading angle of the rectangle, defaults to 0.0
        :param segment_wise: Whether to return segment-wise results, defaults to true
        :param sort: Whether to sort the results, defaults to true
        :return: Polylines within the rotated rectangle
        """

class FlatBush:
    @typing.overload
    def __init__(self) -> None:
        """
        Initialize an empty FlatBush index.
        """
    @typing.overload
    def __init__(self, reserve: int) -> None:
        """
        Initialize a FlatBush index with a reserved capacity.

        :param reserve: Number of items to reserve space for
        """
    @typing.overload
    def add(
        self,
        minX: float,
        minY: float,
        maxX: float,
        maxY: float,
        *,
        label0: int = -1,
        label1: int = -1,
    ) -> int:
        """
        Add a bounding box to the index.

        :param minX: Minimum X coordinate of the bounding box
        :param minY: Minimum Y coordinate of the bounding box
        :param maxX: Maximum X coordinate of the bounding box
        :param maxY: Maximum Y coordinate of the bounding box
        :param label0: First label (optional)
        :param label1: Second label (optional)
        :return: Index of the added item
        """
    @typing.overload
    def add(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        *,
        label0: int = -1,
    ) -> int:
        """
        Add a polyline to the index.

        :param polyline: Polyline coordinates
        :param label0: Label for the polyline (optional)
        :return: Index of the added item
        """
    @typing.overload
    def add(
        self,
        box: numpy.ndarray[numpy.float64[4, 1]],
        *,
        label0: int = -1,
        label1: int = -1,
    ) -> int:
        """
        Add a bounding box to the index using a vector.

        :param box: Vector of [minX, minY, maxX, maxY]
        :param label0: First label (optional)
        :param label1: Second label (optional)
        :return: Index of the added item
        """
    def box(self, index: int) -> numpy.ndarray[numpy.float64[4, 1]]:
        """
        Get the bounding box for a specific index.

        :param index: Index of the item
        :return: Bounding box of the item
        """
    def boxes(
        self,
    ) -> numpy.ndarray[numpy.float64[m, 4], numpy.ndarray.flags.c_contiguous]:
        """
        Get all bounding boxes in the index.

        :return: Reference to the vector of bounding boxes
        """
    def finish(self) -> None:
        """
        Finish the index construction.
        """
    def label(self, index: int) -> numpy.ndarray[numpy.int32[2, 1]]:
        """
        Get the label for a specific index.

        :param index: Index of the item
        :return: Label of the item
        """
    def labels(
        self,
    ) -> numpy.ndarray[numpy.int32[m, 2], numpy.ndarray.flags.c_contiguous]:
        """
        Get all labels in the index.

        :return: Reference to the vector of labels
        """
    def reserve(self, arg0: int) -> None:
        """
        Reserve space for a number of items.

        :param n: Number of items to reserve space for
        """
    @typing.overload
    def search(self, minX: float, minY: float, maxX: float, maxY: float) -> list[int]:
        """
        Search for items within a bounding box.

        :param minX: Minimum X coordinate of the search box
        :param minY: Minimum Y coordinate of the search box
        :param maxX: Maximum X coordinate of the search box
        :param maxY: Maximum Y coordinate of the search box
        :return: Vector of indices of items within the search box
        """
    @typing.overload
    def search(self, bbox: numpy.ndarray[numpy.float64[4, 1]]) -> list[int]:
        """
        Search for items within a bounding box using a vector.

        :param bbox: Vector of [minX, minY, maxX, maxY]
        :return: Vector of indices of items within the search box
        """
    @typing.overload
    def search(
        self,
        min: numpy.ndarray[numpy.float64[2, 1]],
        max: numpy.ndarray[numpy.float64[2, 1]],
    ) -> list[int]:
        """
        Search for items within a bounding box using min and max vectors.

        :param min: Vector of [minX, minY]
        :param max: Vector of [maxX, maxY]
        :return: Vector of indices of items within the search box
        """
    def size(self) -> int:
        """
        Get the number of items in the index.

        :return: Number of items in the index
        """

class KdQuiver(Quiver):
    @staticmethod
    def _filter(
        *,
        arrows: list[Arrow],
        arrow: Arrow,
        params: Quiver.FilterParams,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.int32[m, 1]]:
        """
        Filter arrows based on the given parameters
        """
    @typing.overload
    def __init__(self) -> None:
        """
        Default constructor for KdQuiver
        """
    @typing.overload
    def __init__(self, anchor_lla: numpy.ndarray[numpy.float64[3, 1]]) -> None:
        """
        Constructor for KdQuiver with anchor LLA coordinates
        """
    @typing.overload
    def add(self, polyline: numpy.ndarray[numpy.float64[m, 3]], index: int = -1) -> int:
        """
        Add a polyline to the KdQuiver
        """
    @typing.overload
    def add(
        self,
        polyline: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
        index: int = -1,
    ) -> int:
        """
        Add a 2D polyline to the KdQuiver
        """
    @typing.overload
    def arrow(self, point_index: int) -> Arrow:
        """
        Get the arrow at the given point index
        """
    @typing.overload
    def arrow(self, polyline_index: int, segment_index: int) -> Arrow:
        """
        Get the arrow at the given polyline and segment indices
        """
    @typing.overload
    def arrow(self, polyline_index: int, segment_index: int, *, t: float) -> Arrow:
        """
        Get the arrow at the given polyline, segment indices, and t parameter
        """
    @typing.overload
    def arrow(self, polyline_index: int, *, range: float) -> Arrow:
        """
        Get the arrow at the given polyline index and range
        """
    def arrows(self, indexes: numpy.ndarray[numpy.int32[m, 1]]) -> list[Arrow]:
        """
        Get arrows for the given indexes
        """
    def directions(
        self, indexes: numpy.ndarray[numpy.int32[m, 1]]
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Get directions for the given indexes
        """
    @typing.overload
    def filter(
        self,
        *,
        hits: numpy.ndarray[numpy.int32[m, 1]],
        arrow: Arrow,
        params: Quiver.FilterParams,
    ) -> numpy.ndarray[numpy.int32[m, 1]]:
        """
        Filter hits based on the given parameters
        """
    @typing.overload
    def filter(
        self,
        *,
        hits: numpy.ndarray[numpy.int32[m, 1]],
        norms: numpy.ndarray[numpy.float64[m, 1]],
        arrow: Arrow,
        params: Quiver.FilterParams,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Filter hits and norms based on the given parameters
        """
    @typing.overload
    def index(self, point_index: int) -> numpy.ndarray[numpy.int32[2, 1]]:
        """
        Get the index for the given point index
        """
    @typing.overload
    def index(self, polyline_index: int, segment_index: int) -> int:
        """
        Get the index for the given polyline and segment indices
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        return_squared_l2: bool = False,
    ) -> tuple[int, float]:
        """
        Find the nearest point to the given position
        """
    @typing.overload
    def nearest(
        self, index: int, *, return_squared_l2: bool = False
    ) -> tuple[int, float]:
        """
        Find the nearest point to the point at the given index
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        k: int,
        sort: bool = True,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Find k nearest points to the given position
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        radius: float,
        sort: bool = True,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Find all points within a given radius of the query position
        """
    @typing.overload
    def positions(
        self,
    ) -> numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous]:
        """
        Get all positions in the KdQuiver
        """
    @typing.overload
    def positions(
        self, indexes: numpy.ndarray[numpy.int32[m, 1]]
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Get positions for the given indexes
        """
    def reset(self) -> None:
        """
        Reset the KdQuiver
        """

class KdTree:
    @typing.overload
    def __init__(self, leafsize: int = 10) -> None:
        """
        Initialize KdTree with specified leaf size.

        :param leafsize: Maximum number of points in leaf node, defaults to 10
        """
    @typing.overload
    def __init__(self, points: numpy.ndarray[numpy.float64[m, 3]]) -> None:
        """
        Initialize KdTree with 3D points.

        :param points: 3D points to initialize the tree
        """
    @typing.overload
    def __init__(
        self,
        points: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    ) -> None:
        """
        Initialize KdTree with 2D points.

        :param points: 2D points to initialize the tree
        """
    @typing.overload
    def add(self, points: numpy.ndarray[numpy.float64[m, 3]]) -> None:
        """
        Add 3D points to the KdTree.

        :param points: 3D points to add
        """
    @typing.overload
    def add(
        self,
        points: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    ) -> None:
        """
        Add 2D points to the KdTree.

        :param points: 2D points to add
        """
    def build_index(self, force_rebuild: bool = False) -> None:
        """
        Build the KdTree index.

        :param force_rebuild: Force rebuilding the index even if already built, defaults to false
        """
    def leafsize(self) -> int:
        """
        Get the current leaf size of the KdTree.

        :return: Current leaf size
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        return_squared_l2: bool = False,
    ) -> tuple[int, float]:
        """
        Find the nearest point to the given position.

        :param position: Query position
        :param return_squared_l2: If true, return squared L2 distance, defaults to false
        :return: Tuple of (index, distance)
        """
    @typing.overload
    def nearest(
        self, index: int, *, return_squared_l2: bool = False
    ) -> tuple[int, float]:
        """
        Find the nearest point to the point at the given index.

        :param index: Index of the query point
        :param return_squared_l2: If true, return squared L2 distance, defaults to false
        :return: Tuple of (index, distance)
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        k: int,
        sort: bool = True,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Find k nearest points to the given position.

        :param position: Query position
        :param k: Number of nearest neighbors to find
        :param sort: If true, sort results by distance, defaults to true
        :param return_squared_l2: If true, return squared L2 distances, defaults to false
        :return: Tuple of (indices, distances)
        """
    @typing.overload
    def nearest(
        self,
        position: numpy.ndarray[numpy.float64[3, 1]],
        *,
        radius: float,
        sort: bool = True,
        return_squared_l2: bool = False,
    ) -> tuple[numpy.ndarray[numpy.int32[m, 1]], numpy.ndarray[numpy.float64[m, 1]]]:
        """
        Find all points within a given radius of the query position.

        :param position: Query position
        :param radius: Search radius
        :param sort: If true, sort results by distance, defaults to true
        :param return_squared_l2: If true, return squared L2 distances, defaults to false
        :return: Tuple of (indices, distances)
        """
    def points(
        self,
    ) -> numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous]:
        """
        Get the points in the KdTree.

        :return: Reference to the points in the tree
        """
    def reset(self) -> None:
        """
        Reset the KdTree, clearing all points.
        """
    def reset_index(self) -> None:
        """
        Reset the index of the KdTree.
        """
    def set_leafsize(self, value: int) -> None:
        """
        Set the leaf size of the KdTree.

        :param value: New leaf size value
        """

class LineSegment:
    def __init__(
        self,
        A: numpy.ndarray[numpy.float64[3, 1]],
        B: numpy.ndarray[numpy.float64[3, 1]],
    ) -> None:
        """
        Initialize a LineSegment with two 3D points.
        """
    def distance(self, P: numpy.ndarray[numpy.float64[3, 1]]) -> float:
        """
        Calculate the distance from a point to the line segment.
        """
    def distance2(self, P: numpy.ndarray[numpy.float64[3, 1]]) -> float:
        """
        Calculate the squared distance from a point to the line segment.
        """
    def intersects(
        self, other: LineSegment
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], float, float, float] | None:
        """
        Check if this line segment intersects with another.
        """
    @property
    def A(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the start point of the line segment.
        """
    @property
    def AB(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the vector from A to B.
        """
    @property
    def B(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the end point of the line segment.
        """
    @property
    def length(self) -> float:
        """
        Get the length of the line segment.
        """
    @property
    def length2(self) -> float:
        """
        Get the squared length of the line segment.
        """

class PolylineRuler:
    @staticmethod
    def _along(
        line: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        dist: float,
        *,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Find a point at a specified distance along a polyline.
        """
    @staticmethod
    def _dirs(
        polyline: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Calculate direction vectors for each segment of a polyline.
        """
    @staticmethod
    def _distance(
        a: numpy.ndarray[numpy.float64[3, 1]],
        b: numpy.ndarray[numpy.float64[3, 1]],
        *,
        is_wgs84: bool = False,
    ) -> float:
        """
        Calculate the distance between two points.
        """
    @staticmethod
    def _interpolate(
        A: numpy.ndarray[numpy.float64[3, 1]],
        B: numpy.ndarray[numpy.float64[3, 1]],
        *,
        t: float,
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Interpolate between two points.
        """
    @staticmethod
    def _lineDistance(
        line: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> float:
        """
        Calculate the total length of a polyline.
        """
    @staticmethod
    def _lineSlice(
        start: numpy.ndarray[numpy.float64[3, 1]],
        stop: numpy.ndarray[numpy.float64[3, 1]],
        line: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Extract a portion of a polyline between two points.
        """
    @staticmethod
    def _lineSliceAlong(
        start: float,
        stop: float,
        line: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Extract a portion of a polyline between two distances along it.
        """
    @staticmethod
    def _pointOnLine(
        line: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        P: numpy.ndarray[numpy.float64[3, 1]],
        *,
        is_wgs84: bool = False,
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], int, float]:
        """
        Find the closest point on a polyline to a given point.
        """
    @staticmethod
    def _pointToSegmentDistance(
        P: numpy.ndarray[numpy.float64[3, 1]],
        A: numpy.ndarray[numpy.float64[3, 1]],
        B: numpy.ndarray[numpy.float64[3, 1]],
        *,
        is_wgs84: bool = False,
    ) -> float:
        """
        Calculate the distance from a point to a line segment.
        """
    @staticmethod
    def _ranges(
        polyline: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Calculate cumulative distances along a polyline.
        """
    @staticmethod
    def _squareDistance(
        a: numpy.ndarray[numpy.float64[3, 1]],
        b: numpy.ndarray[numpy.float64[3, 1]],
        *,
        is_wgs84: bool = False,
    ) -> float:
        """
        Calculate the squared distance between two points.
        """
    def N(self) -> int:
        """
        Get the number of points in the polyline.
        """
    def __init__(
        self,
        coords: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
        *,
        is_wgs84: bool = False,
    ) -> None:
        """
        Initialize a PolylineRuler with coordinates and coordinate system.
        """
    def along(self, dist: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Find a point at a specified distance along the polyline.
        """
    @typing.overload
    def arrow(
        self, *, index: int, t: float
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], numpy.ndarray[numpy.float64[3, 1]]]:
        """
        Get the arrow (point and direction) at a specific segment index and interpolation factor.
        """
    @typing.overload
    def arrow(
        self, range: float, *, smooth_joint: bool = True
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], numpy.ndarray[numpy.float64[3, 1]]]:
        """
        Get the arrow (point and direction) at a specific cumulative distance.
        """
    @typing.overload
    def arrows(
        self, ranges: numpy.ndarray[numpy.float64[m, 1]], *, smooth_joint: bool = True
    ) -> tuple[
        numpy.ndarray[numpy.float64[m, 1]],
        numpy.ndarray[numpy.float64[m, 3]],
        numpy.ndarray[numpy.float64[m, 3]],
    ]:
        """
        Get arrows (points and directions) at multiple cumulative distances.
        """
    @typing.overload
    def arrows(
        self, step: float, *, with_last: bool = True, smooth_joint: bool = True
    ) -> tuple[
        numpy.ndarray[numpy.float64[m, 1]],
        numpy.ndarray[numpy.float64[m, 3]],
        numpy.ndarray[numpy.float64[m, 3]],
    ]:
        """
        Get arrows (points and directions) at regular intervals along the polyline.
        """
    @typing.overload
    def at(self, *, range: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the point on the polyline at a specific cumulative distance.
        """
    @typing.overload
    def at(self, *, segment_index: int) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the point on the polyline at a specific segment index.
        """
    @typing.overload
    def at(self, *, segment_index: int, t: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the point on the polyline at a specific segment index and interpolation factor.
        """
    @typing.overload
    def dir(self, *, point_index: int) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the direction vector at a specific point index.
        """
    @typing.overload
    def dir(
        self, *, range: float, smooth_joint: bool = True
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the direction vector at a specific cumulative distance.
        """
    def dirs(self) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Get direction vectors for each segment of the polyline.
        """
    def extended_along(self, range: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the extended cumulative distance along the polyline.
        """
    def is_wgs84(self) -> bool:
        """
        Check if the coordinate system is WGS84.
        """
    def k(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the scale factor for distance calculations.
        """
    def length(self) -> float:
        """
        Get the total length of the polyline.
        """
    def lineDistance(self) -> float:
        """
        Get the total length of the polyline.
        """
    def lineSlice(
        self,
        start: numpy.ndarray[numpy.float64[3, 1]],
        stop: numpy.ndarray[numpy.float64[3, 1]],
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Extract a portion of the polyline between two points.
        """
    def lineSliceAlong(
        self, start: float, stop: float
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Extract a portion of the polyline between two distances along it.
        """
    def local_frame(
        self, range: float, *, smooth_joint: bool = True
    ) -> numpy.ndarray[numpy.float64[4, 4]]:
        """
        Get the local coordinate frame at a specific cumulative distance.
        """
    def pointOnLine(
        self, P: numpy.ndarray[numpy.float64[3, 1]]
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], int, float]:
        """
        Find the closest point on the polyline to a given point.
        """
    def polyline(self) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Get the polyline coordinates.
        """
    @typing.overload
    def range(self, segment_index: int) -> float:
        """
        Get the cumulative distance at a specific segment index.
        """
    @typing.overload
    def range(self, *, segment_index: int, t: float) -> float:
        """
        Get the cumulative distance at a specific segment index and interpolation factor.
        """
    def ranges(self) -> numpy.ndarray[numpy.float64[m, 1]]:
        """
        Get cumulative distances along the polyline.
        """
    def scanline(
        self, range: float, *, min: float, max: float, smooth_joint: bool = True
    ) -> tuple[numpy.ndarray[numpy.float64[3, 1]], numpy.ndarray[numpy.float64[3, 1]]]:
        """
        Generate a scanline perpendicular to the polyline at a specific cumulative distance.
        """
    def segment_index(self, range: float) -> int:
        """
        Get the segment index for a given cumulative distance.
        """
    def segment_index_t(self, range: float) -> tuple[int, float]:
        """
        Get the segment index and interpolation factor for a given cumulative distance.
        """

class Quiver:
    class FilterParams:
        def __init__(self) -> None:
            """
            Default constructor for FilterParams
            """
        @typing.overload
        def angle_slots(self) -> numpy.ndarray[numpy.float64[m, 1]] | None:
            """
            Get the angle slots of the FilterParams
            """
        @typing.overload
        def angle_slots(
            self, arg0: numpy.ndarray[numpy.float64[m, 1]] | None
        ) -> Quiver.FilterParams:
            """
            Set the angle slots of the FilterParams
            """
        def is_trivial(self) -> bool:
            """
            Check if the FilterParams is trivial
            """
        @typing.overload
        def x_slots(self) -> numpy.ndarray[numpy.float64[m, 1]] | None:
            """
            Get the x slots of the FilterParams
            """
        @typing.overload
        def x_slots(
            self, arg0: numpy.ndarray[numpy.float64[m, 1]] | None
        ) -> Quiver.FilterParams:
            """
            Set the x slots of the FilterParams
            """
        @typing.overload
        def y_slots(self) -> numpy.ndarray[numpy.float64[m, 1]] | None:
            """
            Get the y slots of the FilterParams
            """
        @typing.overload
        def y_slots(
            self, arg0: numpy.ndarray[numpy.float64[m, 1]] | None
        ) -> Quiver.FilterParams:
            """
            Set the y slots of the FilterParams
            """
        @typing.overload
        def z_slots(self) -> numpy.ndarray[numpy.float64[m, 1]] | None:
            """
            Get the z slots of the FilterParams
            """
        @typing.overload
        def z_slots(
            self, arg0: numpy.ndarray[numpy.float64[m, 1]] | None
        ) -> Quiver.FilterParams:
            """
            Set the z slots of the FilterParams
            """

    @staticmethod
    def _k(arg0: float) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the constant k
        """
    @typing.overload
    def __init__(self) -> None:
        """
        Default constructor for Quiver
        """
    @typing.overload
    def __init__(self, anchor_lla: numpy.ndarray[numpy.float64[3, 1]]) -> None:
        """
        Constructor for Quiver with anchor LLA coordinates
        """
    def anchor(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the anchor point of the Quiver
        """
    @typing.overload
    def enu2lla(
        self, coords: numpy.ndarray[numpy.float64[3, 1]]
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Convert ENU coordinates to LLA
        """
    @typing.overload
    def enu2lla(
        self, coords: numpy.ndarray[numpy.float64[m, 3]]
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Convert multiple ENU coordinates to LLA
        """
    def forwards(self, arrow: Arrow, delta_x: float) -> Arrow:
        """
        Move the Arrow forward by delta_x
        """
    def inv_k(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the inverse k value of the Quiver
        """
    def is_wgs84(self) -> bool:
        """
        Check if the Quiver is using WGS84 coordinates
        """
    def k(self) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Get the k value of the Quiver
        """
    def leftwards(self, arrow: Arrow, delta_y: float) -> Arrow:
        """
        Move the Arrow leftward by delta_y
        """
    @typing.overload
    def lla2enu(
        self, coords: numpy.ndarray[numpy.float64[3, 1]]
    ) -> numpy.ndarray[numpy.float64[3, 1]]:
        """
        Convert LLA coordinates to ENU
        """
    @typing.overload
    def lla2enu(
        self, coords: numpy.ndarray[numpy.float64[m, 3]]
    ) -> numpy.ndarray[numpy.float64[m, 3]]:
        """
        Convert multiple LLA coordinates to ENU
        """
    def towards(
        self,
        arrow: Arrow,
        delta_frenet: numpy.ndarray[numpy.float64[3, 1]],
        *,
        update_direction: bool = True,
    ) -> Arrow:
        """
        Move the Arrow in Frenet coordinates
        """
    def update(
        self,
        arrow: Arrow,
        delta_enu: numpy.ndarray[numpy.float64[3, 1]],
        *,
        update_direction: bool = True,
    ) -> Arrow:
        """
        Update the Arrow's position and optionally direction
        """
    def upwards(self, arrow: Arrow, delta_z: float) -> Arrow:
        """
        Move the Arrow upward by delta_z
        """

def densify_polyline(
    polyline: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
    *,
    max_gap: float,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    densify polyline, interpolate to satisfy max_gap
    """

@typing.overload
def douglas_simplify(
    coords: numpy.ndarray[numpy.float64[m, 3]],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Simplify a polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def douglas_simplify(
    coords: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.float64[m, 2]]:
    """
    Simplify a 2D polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def douglas_simplify_indexes(
    coords: numpy.ndarray[numpy.float64[m, 3]],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.int32[m, 1]]:
    """
    Get indexes of points to keep when simplifying a polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def douglas_simplify_indexes(
    coords: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.int32[m, 1]]:
    """
    Get indexes of points to keep when simplifying a 2D polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def douglas_simplify_mask(
    coords: numpy.ndarray[numpy.float64[m, 3]],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.int32[m, 1]]:
    """
    Get a mask of points to keep when simplifying a polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def douglas_simplify_mask(
    coords: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    epsilon: float,
    *,
    is_wgs84: bool = False,
    recursive: bool = True,
) -> numpy.ndarray[numpy.int32[m, 1]]:
    """
    Get a mask of points to keep when simplifying a 2D polyline using the Douglas-Peucker algorithm.
    """

@typing.overload
def intersect_segments(
    a1: numpy.ndarray[numpy.float64[2, 1]],
    a2: numpy.ndarray[numpy.float64[2, 1]],
    b1: numpy.ndarray[numpy.float64[2, 1]],
    b2: numpy.ndarray[numpy.float64[2, 1]],
) -> tuple[numpy.ndarray[numpy.float64[2, 1]], float, float] | None:
    """
    Intersect two 2D line segments.
    """

@typing.overload
def intersect_segments(
    a1: numpy.ndarray[numpy.float64[3, 1]],
    a2: numpy.ndarray[numpy.float64[3, 1]],
    b1: numpy.ndarray[numpy.float64[3, 1]],
    b2: numpy.ndarray[numpy.float64[3, 1]],
) -> tuple[numpy.ndarray[numpy.float64[3, 1]], float, float, float] | None:
    """
    Intersect two 3D line segments.
    """

def point_in_polygon(
    *,
    points: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    polygon: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
) -> numpy.ndarray[numpy.int32[m, 1]]:
    """
    point-in-polygon test, returns 0-1 mask
    """

@typing.overload
def polyline_in_polygon(
    polyline: numpy.ndarray[numpy.float64[m, 3]],
    polygon: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    *,
    fc: FastCrossing,
) -> dict[
    tuple[int, float, float, int, float, float], numpy.ndarray[numpy.float64[m, 3]]
]: ...
@typing.overload
def polyline_in_polygon(
    polyline: numpy.ndarray[numpy.float64[m, 3]],
    polygon: numpy.ndarray[numpy.float64[m, 2], numpy.ndarray.flags.c_contiguous],
    *,
    is_wgs84: bool = False,
) -> dict[
    tuple[int, float, float, int, float, float], numpy.ndarray[numpy.float64[m, 3]]
]: ...
def snap_onto_2d(
    P: numpy.ndarray[numpy.float64[2, 1]],
    A: numpy.ndarray[numpy.float64[2, 1]],
    B: numpy.ndarray[numpy.float64[2, 1]],
) -> tuple[numpy.ndarray[numpy.float64[2, 1]], float, float]:
    """
    Snap P onto line segment AB
    """

__version__: str = "0.1.1"



# based on https://raw.githubusercontent.com/marcomusy/vedo/master/examples/basic/spline_tool.py
# recorded video:
#   -   youtube: https://www.youtube.com/watch?v=1dPJ3P84FxE
#   -   bilibil: https://www.bilibili.com/video/BV1D24y1u7uB
from __future__ import annotations

import math
import random

import numpy as np
from vedo import Circle, show  # noqa
from vedo.plotter import Event  # noqa
from vedo.pointcloud import Points  # noqa
from vedo.shapes import Line, Polygon  # noqa

from fast_crossing import PolylineRuler, polyline_in_polygon  # noqa


# https://stackoverflow.com/questions/8997099/algorithm-to-generate-random-2d-polygon
def generate_polygon(
    *,
    center: tuple[float, float] = (0.0, 0.0),
    avg_radius: float = 100.0,
    irregularity: float = 2.0,
    spikiness: float = 0.3,
    num_vertices: int = 100,
    close: bool = True,
) -> list[tuple[float, float]]:
    if irregularity < 0 or irregularity > 1:
        raise ValueError("Irregularity must be between 0 and 1.")
    if spikiness < 0 or spikiness > 1:
        raise ValueError("Spikiness must be between 0 and 1.")
    if close:
        num_vertices -= 1
    irregularity *= 2 * math.pi / num_vertices
    spikiness *= avg_radius

    def random_angle_steps(steps: int, irregularity: float) -> list[float]:
        angles = []
        lower = (2 * math.pi / steps) - irregularity
        upper = (2 * math.pi / steps) + irregularity
        cumsum = 0
        for _ in range(steps):
            angle = random.uniform(lower, upper)
            angles.append(angle)
            cumsum += angle
        cumsum /= 2 * math.pi
        for i in range(steps):
            angles[i] /= cumsum
        return angles

    angle_steps = random_angle_steps(num_vertices, irregularity)
    points = []
    angle = random.uniform(0, 2 * math.pi)

    def clip(value, lower, upper):
        return min(upper, max(value, lower))

    for i in range(num_vertices):
        radius = clip(random.gauss(avg_radius, spikiness), 0, 2 * avg_radius)
        point = (
            center[0] + radius * math.cos(angle),
            center[1] + radius * math.sin(angle),
        )
        points.append(point)
        angle += angle_steps[i]
    if close:
        points.append(points[0])
    return np.array(points)


def on_key_press(evt):
    if evt.keypress == "c":
        print("==== Cleared all points ====", c="r", invert=True)


# def update(self):
#     self.remove([self.spline, self.points])  # remove old points and spline
#     self.points = Points(self.cpoints).ps(10).c("purple5")
#     self.points.pickable(False)  # avoid picking the same point
#     if len(self.cpoints) > 2:
#         self.spline = Spline(self.cpoints, closed=False).c("yellow5").lw(3)
#         self.add(self.points, self.spline)
#     else:
#         self.add(self.points)


radius = 100

polygon = generate_polygon(avg_radius=radius, irregularity=1.0, spikiness=0.2)
polygon = Line(polygon)
# print(polygon.points().shape)

xs = np.linspace(-radius, radius, 30)
ys = np.sin(xs * 2) * radius / 6
rs = np.linspace(0, np.pi, len(xs))[::-1]  # upper half circle
xs += np.cos(rs) * radius
ys += np.sin(rs) * radius
ys -= radius / 3
polyline = Line(np.c_[xs, ys, np.ones(len(xs))])
# print(polyline.points().shape)

sptool = None


def update_polyline_segments():
    if sptool is None:
        coords = polyline.points()
    else:
        coords = sptool.spline().points()
    chunks = polyline_in_polygon(coords, polygon.points()[:, :2])
    layer = []
    np.random.seed(0)
    for label, coords in chunks.items():
        l1, l2 = label[:3], label[3:]  # noqa
        coords[:, 2] += radius / 50
        seg = Line(coords).linewidth(10)
        seg.color((np.random.random(3) * 200 + 55).astype(np.uint8).tolist())
        # length = l2[-1] - l1[-1]
        # ruler = PolylineRuler(coords)
        # ranges = (np.copy(ruler.ranges()) + l1[-1]).round(3).tolist()
        # seg.labels(ranges, ratio=100)
        layer.append(seg)
    return layer


plt = show([polygon], __doc__, interactive=False, axes=1)
sptool = plt.add_spline_tool(polyline, closed=False)
layer = update_polyline_segments()
plt.add(layer)


def on_sptool(obj, evt):
    global layer
    plt.remove(layer)
    layer = update_polyline_segments()
    plt.add(layer)


sptool.AddObserver("InteractionEvent", on_sptool)

plt.interactive()



from __future__ import annotations

import math
import os
import random
import time

import numpy as np
from loguru import logger


def point_in_polygon_matplotlib(points: np.ndarray, polygon: np.ndarray) -> np.ndarray:
    from matplotlib.path import Path

    tic = time.time()
    path = Path(polygon)
    mask = path.contains_points(points).astype(np.int32)
    logger.info(
        f"point_in_polygon_matplotlib, secs:{time.time() - tic:.9f} ({mask.sum():,}/{len(points)})"
    )
    return mask


def point_in_polygon_cubao(points: np.ndarray, polygon: np.ndarray) -> np.ndarray:
    """
    it's migrated from matplotlib
    """
    from fast_crossing import point_in_polygon

    tic = time.time()
    mask = point_in_polygon(points=points, polygon=polygon)
    logger.info(
        f"point_in_polygon_cubao, secs:{time.time() - tic:.9f} ({mask.sum():,}/{len(points)})"
    )
    return mask


def point_in_polygon_polygons(points: np.ndarray, polygon: np.ndarray) -> np.ndarray:
    """
    not ready
    """
    import polygons

    polygon = polygon.tolist()
    num_edges_children = 4
    num_nodes_children = 4
    tree = polygons.build_search_tree(polygon, num_edges_children, num_nodes_children)
    return polygons.points_are_inside(tree, points).astype(np.int32)


def point_in_polygon_shapely(points: np.ndarray, polygon: np.ndarray) -> np.ndarray:
    import shapely
    from shapely.geometry import Polygon

    tic = time.time()
    polygon = Polygon(polygon)
    points = shapely.points(points)
    mask = shapely.contains(polygon, points).astype(np.int32)
    logger.info(
        f"point_in_polygon_shapely, secs:{time.time() - tic:.9f} ({mask.sum():,}/{len(points)})"
    )
    return mask


def load_points(path: str):
    return np.load(path)


def load_polygon(path: str):
    if path.endswith((".npy", ".pcd")):
        load_points(path)


def write_mask(mask: np.ndarray, path: str):
    os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
    np.save(path, mask)
    logger.info(f"wrote to {path}")


def wrapping(fn):
    def wrapped_fn(input_points: str, input_polygon: str, output_path: str):
        points = load_points(input_points)[:, :2]
        polygon = load_polygon(input_polygon)[:, :2]
        mask = fn(points, polygon)
        write_mask(mask, output_path)

    return wrapped_fn


# https://stackoverflow.com/questions/8997099/algorithm-to-generate-random-2d-polygon
def generate_polygon(
    center: tuple[float, float],
    avg_radius: float,
    irregularity: float,
    spikiness: float,
    num_vertices: int,
) -> list[tuple[float, float]]:
    """
    Start with the center of the polygon at center, then creates the
    polygon by sampling points on a circle around the center.
    Random noise is added by varying the angular spacing between
    sequential points, and by varying the radial distance of each
    point from the centre.

    Args:
        center (Tuple[float, float]):
            a pair representing the center of the circumference used
            to generate the polygon.
        avg_radius (float):
            the average radius (distance of each generated vertex to
            the center of the circumference) used to generate points
            with a normal distribution.
        irregularity (float):
            variance of the spacing of the angles between consecutive
            vertices.
        spikiness (float):
            variance of the distance of each vertex to the center of
            the circumference.
        num_vertices (int):
            the number of vertices of the polygon.
    Returns:
        List[Tuple[float, float]]: list of vertices, in CCW order.
    """
    # Parameter check
    if irregularity < 0 or irregularity > 1:
        raise ValueError("Irregularity must be between 0 and 1.")
    if spikiness < 0 or spikiness > 1:
        raise ValueError("Spikiness must be between 0 and 1.")

    irregularity *= 2 * math.pi / num_vertices
    spikiness *= avg_radius
    angle_steps = random_angle_steps(num_vertices, irregularity)

    # now generate the points
    points = []
    angle = random.uniform(0, 2 * math.pi)
    for i in range(num_vertices):
        radius = clip(random.gauss(avg_radius, spikiness), 0, 2 * avg_radius)
        point = (
            center[0] + radius * math.cos(angle),
            center[1] + radius * math.sin(angle),
        )
        points.append(point)
        angle += angle_steps[i]

    return points


def random_angle_steps(steps: int, irregularity: float) -> list[float]:
    """Generates the division of a circumference in random angles.

    Args:
        steps (int):
            the number of angles to generate.
        irregularity (float):
            variance of the spacing of the angles between consecutive vertices.
    Returns:
        List[float]: the list of the random angles.
    """
    # generate n angle steps
    angles = []
    lower = (2 * math.pi / steps) - irregularity
    upper = (2 * math.pi / steps) + irregularity
    cumsum = 0
    for _ in range(steps):
        angle = random.uniform(lower, upper)
        angles.append(angle)
        cumsum += angle

    # normalize the steps so that point 0 and point n+1 are the same
    cumsum /= 2 * math.pi
    for i in range(steps):
        angles[i] /= cumsum
    return angles


def clip(value, lower, upper):
    """
    Given an interval, values outside the interval are clipped to the interval
    edges.
    """
    return min(upper, max(value, lower))


def generate_test_data(
    output_dir: str = ".",
    *,
    label: str = None,
    num: int = 10000,
    width: float = 800,
    height: float = 600,
    radius: float = 250,
):
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)

    if not label:
        label = f"random_num_{num}__bbox_{width:.2f}x{height:.2f}__radius_{radius:.2f}"
    xs = (np.random.random(num) - 0.5) * width
    ys = (np.random.random(num) - 0.5) * height
    xyzs = np.zeros((num, 2))
    xyzs[:, 0] = xs
    xyzs[:, 1] = ys
    path = f"{output_dir}/{label}__points.npy"
    np.save(path, xyzs)
    logger.info(f"wrote to {path}")

    polygon = np.array(
        generate_polygon(
            [0.0, 0.0],
            avg_radius=radius,
            irregularity=1.0,
            spikiness=0.6,
            num_vertices=60,
        )
    )
    path = f"{output_dir}/{label}__polygon.npy"
    np.save(path, polygon)
    logger.info(f"wrote to {path}")


def point_in_polygon_all():
    pass


if __name__ == "__main__":
    import fire

    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(
        {
            "all": point_in_polygon_all,
            "matplotlib": wrapping(point_in_polygon_matplotlib),
            "cubao": wrapping(point_in_polygon_cubao),
            "polygons": wrapping(point_in_polygon_polygons),
            "shapely": wrapping(point_in_polygon_shapely),
            "generate_test_data": generate_test_data,
        }
    )



from __future__ import annotations
import numpy
import typing

__all__ = [
    "R_ecef_enu",
    "T_ecef_enu",
    "apply_transform",
    "apply_transform_inplace",
    "cheap_ruler_k",
    "ecef2enu",
    "ecef2lla",
    "enu2ecef",
    "enu2lla",
    "lla2ecef",
    "lla2enu",
]

def R_ecef_enu(lon: float, lat: float) -> numpy.ndarray[numpy.float64[3, 3]]:
    """
    Get rotation matrix from ECEF to ENU coordinate system.
    """

@typing.overload
def T_ecef_enu(
    lon: float, lat: float, alt: float
) -> numpy.ndarray[numpy.float64[4, 4]]:
    """
    Get transformation matrix from ECEF to ENU coordinate system.
    """

@typing.overload
def T_ecef_enu(
    lla: numpy.ndarray[numpy.float64[3, 1]],
) -> numpy.ndarray[numpy.float64[4, 4]]:
    """
    Get transformation matrix from ECEF to ENU coordinate system using LLA vector.
    """

def apply_transform(
    T: numpy.ndarray[numpy.float64[4, 4]],
    coords: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Apply transformation matrix to coordinates.
    """

def apply_transform_inplace(
    T: numpy.ndarray[numpy.float64[4, 4]],
    coords: numpy.ndarray[
        numpy.float64[m, 3],
        numpy.ndarray.flags.writeable,
        numpy.ndarray.flags.c_contiguous,
    ],
    *,
    batch_size: int = 1000,
) -> None:
    """
    Apply transformation matrix to coordinates in-place.
    """

def cheap_ruler_k(latitude: float) -> numpy.ndarray[numpy.float64[3, 1]]:
    """
    Get the cheap ruler's unit conversion factor for a given latitude.
    """

def ecef2enu(
    ecefs: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
    *,
    anchor_lla: numpy.ndarray[numpy.float64[3, 1]] | None = None,
    cheap_ruler: bool = False,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert ECEF to ENU (East, North, Up) coordinates.
    """

@typing.overload
def ecef2lla(x: float, y: float, z: float) -> numpy.ndarray[numpy.float64[3, 1]]:
    """
    Convert ECEF coordinates to LLA (Longitude, Latitude, Altitude).
    """

@typing.overload
def ecef2lla(
    ecefs: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert multiple ECEF coordinates to LLA (Longitude, Latitude, Altitude).
    """

def enu2ecef(
    enus: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
    *,
    anchor_lla: numpy.ndarray[numpy.float64[3, 1]],
    cheap_ruler: bool = False,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert ENU (East, North, Up) to ECEF coordinates.
    """

def enu2lla(
    enus: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
    *,
    anchor_lla: numpy.ndarray[numpy.float64[3, 1]],
    cheap_ruler: bool = True,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert ENU (East, North, Up) to LLA (Longitude, Latitude, Altitude) coordinates.
    """

@typing.overload
def lla2ecef(lon: float, lat: float, alt: float) -> numpy.ndarray[numpy.float64[3, 1]]:
    """
    Convert LLA (Longitude, Latitude, Altitude) to ECEF coordinates.
    """

@typing.overload
def lla2ecef(
    llas: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert multiple LLA (Longitude, Latitude, Altitude) to ECEF coordinates.
    """

def lla2enu(
    llas: numpy.ndarray[numpy.float64[m, 3], numpy.ndarray.flags.c_contiguous],
    *,
    anchor_lla: numpy.ndarray[numpy.float64[3, 1]] | None = None,
    cheap_ruler: bool = True,
) -> numpy.ndarray[numpy.float64[m, 3]]:
    """
    Convert LLA (Longitude, Latitude, Altitude) to ENU (East, North, Up) coordinates.
    """



from __future__ import annotations

from vedo import Picture, dataurl
from vedo.applications import SplinePlotter  # ready to use class!

pic = Picture(dataurl + "images/embryo.jpg")

plt = SplinePlotter(pic)
plt.show(mode="image", zoom="tight")
print("Npts =", len(plt.cpoints), "NSpline =", plt.line.npoints)



from __future__ import annotations

import os
import sys
import time

import numpy as np
import pytest

from fast_crossing import (
    Arrow,
    FastCrossing,
    FlatBush,
    KdQuiver,
    KdTree,
    PolylineRuler,
    Quiver,
    densify_polyline,
    point_in_polygon,
    polyline_in_polygon,
    tf,
)


def test_fast_crossing():
    fc = FastCrossing()

    # add your polylines
    """
                    2 C
                    |
                    1 D
    0               |                  5
    A---------------o------------------B
                    |
                    |
                    -2 E
    """
    fc.add_polyline(np.array([[0.0, 0.0], [5.0, 0.0]]))  # AB
    fc.add_polyline(np.array([[2.5, 2.0], [2.5, 1.0], [2.5, -2.0]]))  # CDE

    # build index
    fc.finish()

    # num_polylines
    assert fc.num_polylines() == 2
    rulers = fc.polyline_rulers()
    assert len(rulers) == 2
    ruler0 = fc.polyline_ruler(0)
    ruler1 = fc.polyline_ruler(1)
    assert not ruler0.is_wgs84()
    assert not ruler1.is_wgs84()
    assert ruler0.length() == 5
    assert ruler1.length() == 4
    assert fc.polyline_ruler(10) is None

    # intersections
    ret = fc.intersections([1.5, 0], [3.5, 2])
    assert len(ret) == 2
    assert np.linalg.norm(fc.coordinates(ret[0]) - [1.5, 0, 0]) < 1e-15
    assert np.linalg.norm(fc.coordinates(ret[1]) - [2.5, 1, 0]) < 1e-15
    xyz = fc.coordinates(0, 0, 0.2)
    assert np.linalg.norm(xyz - [1.0, 0, 0]) < 1e-15
    with pytest.raises(IndexError) as excinfo:
        xyz = fc.coordinates(2, 0, 0.5)
    assert "map::at" in str(excinfo)

    xyz, dir = fc.arrow(polyline_index=0, point_index=0)
    assert np.all(xyz == [0, 0, 0])
    assert np.all(dir == [1, 0, 0])

    # query all line segment intersections
    # [
    #    (array([2.5, 0. ]),
    #     array([0.5       , 0.33333333]),
    #     array([0, 0], dtype=int32),
    #     array([1, 1], dtype=int32))
    # ]
    ret = fc.intersections()
    # print(ret)
    assert len(ret) == 1
    for xy, ts, label1, label2 in ret:
        assert np.all(xy == [2.5, 0])
        assert np.all(ts == [0.5, 1 / 3.0])
        assert np.all(label1 == [0, 0])
        assert np.all(label2 == [1, 1])

    # query intersections against provided polyline
    polyline = np.array([[-6.0, -1.0], [-5.0, 1.0], [5.0, -1.0]])
    ret = fc.intersections(polyline)
    ret = np.array(ret)
    xy = ret[:, 0]
    ts = ret[:, 1]
    label1 = ret[:, 2]
    label2 = ret[:, 3]
    # print(ret, xy, ts, label1, label2)
    assert np.all(xy[0] == [0, 0])
    assert np.all(xy[1] == [2.5, -0.5])
    assert np.all(ts[0] == [0.5, 0])
    assert np.all(ts[1] == [0.75, 0.5])
    assert np.all(label1 == [[0, 1], [0, 1]])
    assert np.all(label2 == [[0, 0], [1, 1]])

    polyline2 = np.column_stack((polyline, np.zeros(len(polyline))))
    ret2 = np.array(fc.intersections(polyline2[:, :2]))
    assert str(ret) == str(ret2)


def test_fast_crossing_intersection3d():
    fc = FastCrossing()
    """
                    2 C
                    |
                    1 D
    0               |                  5
    A---------------o------------------B
                    |
                    |
                    -2 E
    """
    fc.add_polyline(np.array([[0.0, 0.0, 0.0], [5.0, 0.0, 100]]))  # AB
    fc.add_polyline(np.array([[2.5, 2.0, 0.0], [2.5, 1.0, 100], [2.5, -2.0, 0]]))  # CDE
    fc.finish()
    ret = fc.intersections()
    assert len(ret) == 1
    ret = ret[0]
    xyz1 = fc.coordinates(ret, second=False)
    xyz2 = fc.coordinates(ret)
    assert np.linalg.norm(xyz1 - [2.5, 0, 50]) < 1e-10
    assert np.linalg.norm(xyz2 - [2.5, 0, 2 / 3 * 100.0]) < 1e-10


def test_fast_crossing_auto_rebuild_flatbush():
    fc = FastCrossing()
    fc.add_polyline(np.array([[0.0, 0.0, 0.0], [5.0, 0.0, 100]]))  # AB
    fc.add_polyline(np.array([[2.5, 2.0, 0.0], [2.5, 1.0, 100], [2.5, -2.0, 0]]))  # CDE
    ret = fc.intersections()
    assert len(ret) == 1

    fc.add_polyline([[1.5, 0], [3.5, 2]])
    ret = fc.intersections()
    assert len(ret) == 4  # should dedup to 3?


def test_fast_crossing_filter_by_z():
    fc = FastCrossing()
    fc.add_polyline([[0, 0, 0], [1, 0, 0]])
    fc.add_polyline([[0, 0, 10], [1, 0, 10]])
    fc.add_polyline([[0, 0, 20], [1, 0, 20]])
    ret = fc.intersections([[0.5, -1], [0.5, 1]])
    assert len(ret) == 3

    ret = fc.intersections([[0.5, -1], [0.5, 1]], z_min=-1, z_max=1)
    assert len(ret) == 1
    assert fc.coordinates(ret[0])[2] == 0

    ret = fc.intersections([[0.5, -1, 10.5], [0.5, 1, 10.5]], z_min=-1, z_max=1)
    assert len(ret) == 1
    assert np.all(fc.coordinates(ret[0]) == [0.5, 0, 10])
    assert np.all(fc.coordinates(ret[0], second=True) == [0.5, 0, 10])
    # don't do this, first point is not in fc!!!
    # assert np.all(fc.coordinates(ret[0], second=False) == [0.5, 0, 10.5])

    ret = fc.intersections([[0.5, -1, 20], [0.5, 1, 20]], z_min=-1, z_max=1)
    assert len(ret) == 1
    assert fc.coordinates(ret[0])[2] == 20

    ret = fc.intersections([[0.5, -1, 15], [0.5, 1, 15]], z_min=-6, z_max=6)
    assert len(ret) == 2
    assert fc.coordinates(ret[0])[2] == 10
    assert fc.coordinates(ret[1])[2] == 20

    assert len(fc.intersections()) == 3  # 3 == C(3,2): choose 2 from 3
    # for row in fc.intersections():
    #     print(row)


def test_fast_crossing_dedup():
    # should be stable
    for _ in range(100):
        fc = FastCrossing()
        fc.add_polyline([[0, 0, 0], [1, 0, 0], [2, 0, 0]])
        fc.add_polyline([[0, 1, 0], [1, 1, 0], [2, 1, 0]])

        ret = fc.intersections([[1, -1], [1, 1]])
        assert len(ret) == 2
        assert np.all(ret[0][-1] == [0, 0]), ret
        assert np.all(ret[1][-1] == [1, 0]), ret
        assert ret[0][1][1] == 1.0, ret
        assert ret[1][1][1] == 1.0, ret

        ret = fc.intersections([[1, -1], [1, 1]], dedup=False)
        # for idx, row in enumerate(ret):
        #     print(idx, row)
        assert len(ret) == 4


def test_fast_crossing_single_polyline():
    fc = FastCrossing()
    fc.add_polyline([[0, 0, 0], [1, 0, 0], [2, 0, 0]])
    assert not fc.intersections()


def test_fast_crossing_single_polyline_self_intersection():
    fc = FastCrossing()
    fc.add_polyline([[0, 0, 0], [1, 0, 0], [1, 1, 0], [-1, -1, 0]])
    assert len(fc.intersections()) == 1
    assert not fc.intersections(z_offset_range=[0.0, 10.0], self_intersection=0)
    assert fc.intersections(z_offset_range=[0.0, 0.0])
    assert not fc.intersections(z_offset_range=[1e-15, 1.0])

    fc = FastCrossing()
    fc.add_polyline([[0, 0, 0], [1, 0, 0], [1, 1, 0], [-1, -1, 0]])
    fc.add_polyline(np.array([[0.0, 0.0], [5.0, 0.0]]))  # AB
    fc.add_polyline(np.array([[2.5, 2.0], [2.5, 1.0], [2.5, -2.0]]))  # CDE
    self_inter = fc.intersections(z_offset_range=[-1.0, 1e10], self_intersection=1)
    assert len(self_inter) == 1


def test_densify():
    coords = np.array(
        [
            [0, 0, 0],
            [2, 0, 0],
            [5, 0, 0],
        ],
        dtype=np.float64,
    )
    dense = densify_polyline(coords, max_gap=1.0)
    assert len(dense) == 6
    dense = densify_polyline(coords, max_gap=2.0)
    assert len(dense) == 4
    dense = densify_polyline(coords, max_gap=3.0)
    assert len(dense) == 3
    assert len(densify_polyline(coords, max_gap=3.0 + 1e-3)) == 3
    assert len(densify_polyline(coords, max_gap=3.0 - 1e-3)) == 4


def test_point_in_polygon():
    polygon = [[0, 0], [1, 0], [1, 1], [0, 1]]
    points = [[0.5, 0.5], [10, 0]]
    mask = point_in_polygon(points=points, polygon=polygon)
    assert np.all(mask == [1, 0])


def test_kdtree():
    xs = np.linspace(0, 10, 101)
    xyzs = np.zeros((len(xs), 3))
    xyzs[:, 0] = xs
    tree = KdTree(xyzs)
    assert tree.points().shape == (101, 3)
    idx, dist = tree.nearest(0)
    assert idx == 1, dist == 0.1
    idx, dist = tree.nearest([0.0, 0.0, 0.0])
    assert idx == 0, dist == 0.0
    idx, dist = tree.nearest([0.0, 0.0, 0.0], k=4)
    assert np.all(idx == [0, 1, 2, 3])
    np.testing.assert_allclose(dist, [0.0, 0.1, 0.2, 0.3], atol=1e-15)
    idx, dist = tree.nearest([0.0, 0.0, 0.0], radius=0.25)
    assert np.all(idx == [0, 1, 2])
    np.testing.assert_allclose(dist, [0.0, 0.1, 0.2], atol=1e-15)


def _test_cKDTree_query(KDTree):
    x, y = np.mgrid[0:5, 2:8]
    tree = KDTree(np.c_[x.ravel(), y.ravel()])

    expected_k1 = [[2.0, 0.2236068], [0, 13]]
    expected_k2 = [[[2.0, 2.23606798], [0.2236068, 0.80622577]], [[0, 6], [13, 19]]]
    for k, expected in zip([1, 2], [expected_k1, expected_k2]):
        dd, ii = tree.query([[0, 0], [2.2, 2.9]], k=k)
        DD, II = expected
        np.testing.assert_allclose(dd, DD, atol=1e-6)
        np.testing.assert_allclose(ii, II, atol=1e-6)

    expected_k1 = [
        [[2.0], [0.22360679774997916]],
        [[0], [13]],
    ]
    expected_k2 = [
        [[2.23606797749979], [0.8062257748298548]],
        [[6], [19]],
    ]
    expected_k1_k2 = [
        [[2.0, 2.23606797749979], [0.22360679774997916, 0.8062257748298548]],
        [[0, 6], [13, 19]],
    ]
    for k, expected in zip(
        [[1], [2], [1, 2]], [expected_k1, expected_k2, expected_k1_k2]
    ):
        dd, ii = tree.query([[0, 0], [2.2, 2.9]], k=k)
        DD, II = expected
        np.testing.assert_allclose(dd, DD, atol=1e-6)
        np.testing.assert_allclose(ii, II, atol=1e-6)

    x, y = np.mgrid[0:4, 0:4]
    points = np.c_[x.ravel(), y.ravel()]
    tree = KDTree(points)
    ret = sorted(tree.query_ball_point([2 + 1e-15, 1e-15], 1 + 1e-9))
    assert np.all(np.array(ret) == [4, 8, 9, 12])

    ret = tree.query_ball_point([[2, 0], [3, 0]], 1 + 1e-9)
    assert np.all(np.array(sorted(ret[0])) == [4, 8, 9, 12])
    assert np.all(np.array(sorted(ret[1])) == [8, 12, 13])

    ret = tree.query_ball_point([[2, 0], [3, 0]], 1 + 1e-9, return_length=True)
    assert np.all(np.array(ret) == [4, 3])


def test_scipy_cKDTree():
    from scipy.spatial import cKDTree

    _test_cKDTree_query(cKDTree)


def test_nanoflann_KDTree():
    from fast_crossing.spatial import KDTree

    _test_cKDTree_query(KDTree)


def test_arrow():
    arrow = Arrow()
    assert np.all(arrow.label() == [-1, -1])
    assert np.isnan(arrow.t())
    assert np.isnan(arrow.range())
    assert not arrow.has_index()
    assert np.all(arrow.position() == [0, 0, 0])
    assert np.all(arrow.direction() == [1, 0, 0])
    h = arrow.heading()
    assert h == 90.0

    arrow.label([2, 3]).t(0.5).range(23.0)
    assert arrow.has_index()

    arrow.position([1, 2, 3])
    assert np.all(arrow.position() == [1, 2, 3])
    arrow.direction([3, 4, 12])
    assert np.all(arrow.direction() == [3, 4, 12])
    arrow.direction(Arrow._unit_vector([3, 4, 0]))
    np.testing.assert_allclose(arrow.direction(), [3 / 5, 4 / 5, 0], atol=1e-6)

    arrow.label([5, 10])
    assert np.all(arrow.label() == [5, 10])
    arrow.label(5, 20)
    assert np.all(arrow.label() == [5, 20])
    arrow.label(5, 20, t=0.3, range=23.0)
    assert arrow.t() == 0.3
    assert arrow.range() == 23.0
    arrow.label(5, 20, range=46.0)
    assert arrow.t() == 0.3
    assert arrow.range() == 46.0

    arrow.heading(90.0)
    np.testing.assert_allclose(arrow.direction(), [1, 0, 0], atol=1e-8)
    np.testing.assert_allclose(arrow.heading(), 90.0, atol=1e-8)
    arrow.heading(0.0)
    np.testing.assert_allclose(arrow.direction(), [0, 1, 0], atol=1e-8)
    np.testing.assert_allclose(arrow.heading(), 0.0, atol=1e-8)

    arrow.t(0.000001)
    assert "label:(5/20/0.000)" in repr(arrow)
    # same like f-string format
    assert f"{0.009:.2f}" == "0.01"
    assert f"{0.001:.2f}" == "0.00"

    arrow2 = arrow.copy()
    arrow2.t(0.3)
    assert arrow.t() == 0.000001

    arrow = Arrow().label([0, 0]).t(0)
    assert not arrow.has_index()
    assert arrow.has_index(check_range=False)
    arrow.range(10)
    assert arrow.has_index()
    assert arrow.has_index(check_range=False)
    arrow.reset_index()
    assert not arrow.has_index()


def test_quiver():
    quiver = Quiver()
    assert not quiver.is_wgs84()
    assert np.all(quiver.k() == [1, 1, 1])
    assert np.all(quiver.anchor() == [0, 0, 0])

    quiver = Quiver([123, 45, 6])
    assert quiver.is_wgs84()
    assert np.all(quiver.anchor() == [123, 45, 6])
    k = quiver.k()
    np.testing.assert_allclose(k, [78846.8350939781, 111131.7774141756, 1.0], atol=1e-6)
    np.testing.assert_allclose(np.sum(k @ quiver.inv_k()), 3.0, atol=1e-16)

    arrow = Arrow([123, 45, 6])
    updated = quiver.update(arrow, [1, 1, 1], update_direction=True)
    np.testing.assert_allclose(
        updated.position(), [123.00001268281724, 45.000008998326344, 7], atol=1e-8
    )
    np.testing.assert_allclose(updated.heading(), 45, atol=1e-8)

    updated = quiver.update(arrow, [1, 1, 0], update_direction=False)
    np.testing.assert_allclose(updated.direction(), [1, 0, 0], atol=1e-8)

    # delta
    #   ^ y (north)
    #   |
    #   @->
    #   |
    #   o---------> x (east)
    #
    quiver = Quiver()
    arrow = Arrow([0, 1, 0], direction=[1, 0, 0])
    # towards (delta in Frenet, x->forwards, y->leftwards, z->upwards)
    updated = quiver.towards(arrow, [2, 0, 0])
    np.testing.assert_allclose(updated.position(), [2, 1, 0], atol=1e-8)
    # update (delta in EUN, x->east, y->north, z->up)
    updated = quiver.update(arrow, [2, 0, 0])
    np.testing.assert_allclose(updated.position(), [2, 1, 0], atol=1e-8)
    arrow = Arrow([0, 1, 0], direction=Arrow._unit_vector([1, 1, 0]))
    updated = quiver.towards(arrow, [3, 3, 0])
    sqrt2 = np.sqrt(2)
    np.testing.assert_allclose(updated.position(), [0, 3 * sqrt2 + 1, 0], atol=1e-8)
    np.testing.assert_allclose(updated.direction(), [0, 1, 0], atol=1e-8)
    updated = quiver.update(arrow, [3, 3, 0])
    np.testing.assert_allclose(updated.position(), [3, 4, 0], atol=1e-8)
    np.testing.assert_allclose(
        updated.direction(), [sqrt2 / 2, sqrt2 / 2, 0], atol=1e-8
    )


def test_kdquiver():
    quiver = KdQuiver()
    """
    #  0   1   2   3   4   5   6  7   8   9   10   11  12  13   14
    #                                                  o
    #                                                 /
    #0 o-------------o-------------------------------+-----------o
    #1        o-------------o-----------------o     /
    #2                   o-------------o----------o
    #3 o-------------o-------------------------------------------o right-to-left
    #4 o-------------o-------------------------------------------o z += 10
    """
    y = 5.0
    quiver.add([[0, y], [3.5, y], [14.0, y]])  # 0
    y -= 1
    quiver.add([[1.8, y], [5.2, y], [10.0, y]])  # 1
    y -= 1
    quiver.add([[4.5, y], [8.2, y], [10.8, y], [12.0, y + 5]])  # 2
    y -= 1
    quiver.add([[0, y], [3.5, y], [14.0, y]][::-1])  # 3
    y -= 1
    quiver.add([[0, y, 10.0], [3.5, y, 10.0], [14.0, y, 10.0]])  # 4

    # nearest
    i, d = quiver.nearest([0.0, 5.0, 0.0])
    a = quiver.arrow(i)
    # print(i, d, a)
    assert i == 0 and d == 0.0 and a.polyline_index() == 0 and a.segment_index() == 0

    i, d = quiver.nearest(0)
    a = quiver.arrow(i)
    # print(i, d, a)
    assert i == 3 and a.polyline_index() == 1 and a.segment_index() == 0

    ii, dd = quiver.nearest([4.0, 2.5, 0.0], radius=3)
    arrows = quiver.arrows(ii)
    # print(ii, dd, arrows)
    # for arrow in arrows:
    #     print(arrow)
    assert len(ii) == len(dd) == len(arrows) == 5
    assert np.all(ii == [6, 11, 4, 1, 3])
    ii2, dd2 = quiver.nearest([4.0, 2.5, 0.0], k=5)
    assert np.all(ii == ii2)
    assert np.all(dd == dd2)

    filter = Quiver.FilterParams()
    assert filter.x_slots() is None
    assert filter.y_slots() is None
    assert filter.z_slots() is None
    assert filter.angle_slots() is None
    filter.x_slots([1, 2, 3, 4]).angle_slots([-50, 50])
    assert np.all(filter.x_slots() == [1, 2, 3, 4])
    assert np.all(filter.angle_slots() == [-50, 50])

    arrow = Arrow([4.0, 2.5, 0.0], [1.0, 0.0, 0.0])
    ii, dd = quiver.nearest(arrow.position(), radius=3)
    assert len(ii) == 5
    # for arr in quiver.arrows(ii):
    #     print('\n', arr, sep=';')
    ii2 = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams(),
    )
    assert np.all(ii == ii2)
    ii_lefts = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().y_slots([0.0, 10]),  # only lefts
    )
    assert len(ii_lefts) == 4
    ii_rights = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().y_slots([-10, 0]),  # only rights
    )
    assert len(ii_rights) == 1
    assert np.all(np.sort(ii) == np.sort([*ii_lefts, *ii_rights]))
    ii, dd = quiver.nearest(arrow.position(), radius=100)
    assert len(ii) == 16
    ii, dd = quiver.nearest(arrow.position(), radius=100)
    ii = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().z_slots([9, 11]),
    )
    assert len(ii) == 3


def test_kdquiver_filter_by_angle():
    quiver = KdQuiver()
    vecs = {}
    headings = [0, 30, 60, 90, 120]
    for heading in headings:
        rad = np.radians(90.0 - heading)
        vec = np.array([np.cos(rad), np.sin(rad), 0.0])
        vecs[heading] = vec
        polyline = np.array([vec * 80, vec * 100])
        quiver.add(polyline, index=heading)
    np.testing.assert_allclose(30, Arrow._angle(vecs[30], ref=vecs[60]), atol=1e-15)
    np.testing.assert_allclose(-30, Arrow._angle(vecs[90], ref=vecs[60]), atol=1e-15)

    arrow = Arrow([0, 0, 0]).heading(60)
    ii, dd = quiver.nearest(arrow.position(), radius=90)
    assert len(ii) == len(headings)
    np.testing.assert_allclose(dd, [80.0] * len(ii), atol=1e-15)

    ii60 = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().angle_slots([-1, 1]),
    )
    hits = np.array([a.heading() for a in quiver.arrows(ii60)])
    np.testing.assert_allclose(hits, [60.0], atol=1e-15)

    ii30_60 = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().angle_slots([-1, 31]),
    )
    hits = np.array([a.heading() for a in quiver.arrows(ii30_60)])
    np.testing.assert_allclose(sorted(hits), [30.0, 60.0], atol=1e-15)

    ii30_60_120 = quiver.filter(
        hits=ii,
        arrow=arrow,
        params=Quiver.FilterParams().angle_slots([-1, 31, -61, -59]),
    )
    hits = np.array([a.heading() for a in quiver.arrows(ii30_60_120)])
    np.testing.assert_allclose(sorted(hits), [30.0, 60.0, 120.0], atol=1e-15)


def test_within():
    fc = FastCrossing()
    """
          0A      ┌────────┐
           ****   │    *2C │
                  │    *   │
    1B  ┌─────────o    *   │
      * │                  │
       *│               **********3D
        *         *        │
        └─*───────*────────┘
           *      *
            *     *4E
    """
    polygon = np.array(
        [[0, 0], [-10, 0], [-10, -10], [10, -10], [10, 10], [0, 10]], dtype=np.float64
    )
    fc.add_polyline([[-8, 9], [-2, 9]])  # 0A
    fc.add_polyline([[-12, -1], [-8, -8], [-5, -15]])  # 1B
    fc.add_polyline([[6, 0], [6, 9]])  # 2C
    fc.add_polyline([[7, -5], [18, -5]])  # 3D
    fc.add_polyline([[0, -5], [0, -15]])  # 4E
    hits = np.array(fc.within(polygon=polygon))
    assert np.all(
        hits
        == [
            [1, 0],
            [1, 1],
            [2, 0],
            [3, 0],
            [4, 0],
        ]
    )
    hits = np.array(fc.within(polygon=polygon, segment_wise=False))
    assert np.all(
        hits
        == [
            [1, 1],
            [2, 0],
            [2, 1],
            [3, 0],
            [4, 0],
        ]
    )

    hits = np.array(fc.within(min=np.array([0.0, 0.0]), max=np.array([10.0, 10.0])))
    assert np.all(hits == [[2, 0]])
    hits = np.array(fc.within(min=np.array([-10.0, -10.0]), max=np.array([10.0, 10.0])))
    assert np.all(
        hits
        == [
            [0, 0],
            [1, 0],
            [1, 1],
            [2, 0],
            [3, 0],
            [4, 0],
        ]
    )

    # fc.add_polyline([[-8, 9], [-2, 9]])  # 0A
    hits = np.array(fc.within(center=np.array([-5.0, 10.0]), width=6.0, height=0.5))
    assert not len(hits)
    hits = np.array(
        fc.within(center=np.array([-5.0, 10.0]), width=6.0, height=0.5, heading=30.0)
    )
    assert np.all(hits == [[0, 0]])


def test_nearst():
    fc = FastCrossing()
    tree = fc.quiver()
    assert tree is None
    bush = fc.bush()
    assert bush is None
    """
          0A      ┌────────┐
           ****   │    *2C │
                  │    *   │
    1B  ┌─────────o    *   │
      * │                  │
       *│               **********3D
        *         *        │
        └─*───────*────────┘
           *      *
            *     *4E
    """
    polygon = np.array(
        [[0, 0], [-10, 0], [-10, -10], [10, -10], [10, 10], [0, 10]], dtype=np.float64
    )
    assert len(polygon)
    fc.add_polyline([[-8, 9], [-2, 9]])  # 0A
    fc.add_polyline([[-12, -1], [-8, -8], [-5, -15]])  # 1B
    fc.add_polyline([[6, 0], [6, 9]])  # 2C
    fc.add_polyline([[7, -5], [18, -5]])  # 3D
    fc.add_polyline([[0, -5], [0, -15]])  # 4E

    # nearest
    idx, dist = fc.nearest(np.array([0.0, 0.0, 0.0]))
    assert np.all(idx == [4, 0]) and np.fabs(dist - 5.0) < 1e-6
    idx, dist = fc.nearest(np.array([0.0, 0.0, 0.0]), return_squared_l2=True)
    assert np.all(idx == [4, 0]) and np.fabs(dist - 25.0) < 1e-6

    # nearest
    idx, dist = fc.nearest(np.array([0, 0]))
    assert np.all(idx == [0, 1]) and np.fabs(dist - 6.0) < 1e-6
    idx, dist = fc.nearest(np.array([0, 1]))
    assert np.all(idx == [0, 0]) and np.fabs(dist - 6.0) < 1e-6

    # nearest k
    idx, dist = fc.nearest(np.array([0.0, 0.0, 0.0]), k=2)
    assert np.all(idx == [[4, 0], [2, 0]])
    np.testing.assert_allclose(dist, [5, 6], atol=1e-15)
    # nearest radius
    idx, dist = fc.nearest(np.array([0.0, 0.0, 0.0]), radius=6 + 1e-3)
    assert np.all(idx == [[4, 0], [2, 0]])
    np.testing.assert_allclose(dist, [5, 6], atol=1e-15)
    idx, dist = fc.nearest(np.array([0.0, 0.0, 0.0]), radius=5 + 1e-3)
    assert np.all(idx == [[4, 0]])
    np.testing.assert_allclose(dist, [5], atol=1e-15)

    arrow = Arrow([-4.0, 0.0, 0.0], [0, -1, 0])
    idx, dist = fc.nearest(arrow.position(), radius=10.0)
    # print(idx, dist)
    assert len(idx) == 5
    idx, dist = fc.nearest(
        arrow.position(),
        radius=10.0,
        filter=[
            arrow.direction(),
            Quiver.FilterParams().angle_slots([-1, 1]),
        ],
    )
    # print(idx, dist)
    assert len(idx) == 1
    assert np.all(idx == [[4, 0]])

    tree = fc.quiver()
    assert tree is not None
    bush = fc.bush(autobuild=False)
    assert bush is None
    bush = fc.bush()
    assert bush is not None


def test_nearst_wgs84():
    fc = FastCrossing()
    fc.add_polyline([[-8, 9], [-2, 9]])  # 0A
    fc.add_polyline([[-12, -1], [-8, -8], [-5, -15]])  # 1B
    fc.add_polyline([[6, 0], [6, 9]])  # 2C
    fc.add_polyline([[7, -5], [18, -5]])  # 3D
    fc.add_polyline([[0, -5], [0, -15]])  # 4E
    arrow = Arrow([-4.0, 0.0, 0.0], [0, -1, 0])
    idx, dist = fc.nearest(arrow.position(), radius=10.0)
    expected_ii = np.array([[4, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
    expected_dd = np.array([6.40312424, 8.06225775, 8.94427191, 9.21954446, 9.8488578])
    assert np.all(idx == expected_ii)
    np.testing.assert_allclose(dist, expected_dd, atol=1e-8)
    assert len(idx) == 5
    idx, dist = fc.nearest(
        arrow.position(),
        radius=10.0,
        filter=[
            arrow.direction(),
            Quiver.FilterParams().angle_slots([-1, 1]),
        ],
    )
    assert len(idx) == 1
    assert np.all(idx == [[4, 0]])

    xyzs0 = fc.quiver().positions()
    # print(xyzs0)

    fc2 = FastCrossing(is_wgs84=True)
    anchor_lla = np.array([123.4, 56.7, 8.9])
    for index, ruler in fc.polyline_rulers().items():
        enus = ruler.polyline()
        llas = tf.enu2lla(enus, anchor_lla=anchor_lla)
        fc2.add_polyline(llas, index=index)
    # print(fc2.quiver().anchor())
    xyzs1 = tf.lla2enu(
        tf.enu2lla(fc2.quiver().positions(), anchor_lla=fc2.quiver().anchor()),
        anchor_lla=anchor_lla,
    )
    # print(xyzs1)
    np.testing.assert_allclose(xyzs0, xyzs1, atol=1e-8)

    position = tf.enu2lla([-4, 0, 0], anchor_lla=anchor_lla).reshape(-1)
    arrow = Arrow(position, [0, -1, 0])
    idx, dist = fc2.nearest(arrow.position(), radius=10.0)
    # precision error??
    # print(idx, dist)
    np.testing.assert_allclose(dist[:4], expected_dd[:4], atol=1e-4)
    # assert len(idx) == 5


def __print_chunks(chunks):
    for cidx, ((seg1, t1, r1, seg2, t2, r2), polyline) in enumerate(chunks.items()):
        print(f"\nchunk index: {cidx}", sep=", ")
        print(f"length: {r2 - r1:.3f}")
        print(f"seg={seg1},t={t1:.3f}, r={r1:.2f}", sep="; ")
        print(f"seg={seg2},t={t2:.3f}, r={r2:.2f}")
        print(polyline.round(2).tolist())


def test_polyline_in_polygon():
    """
           2    4
           *    *
      A   /|   /|
       o---+--/-+--------------o D
       |/  | /  |              |
       /   |/   |              |
      /|   *    *              |
    1* |   3    5              |
      Bo-----------------------o
                               C
    """
    polygon_ABCD = np.array(
        [
            [0.0, 0.0],  # A
            [0.0, -10.0],  # B
            [20.0, -10.0],  # C
            [20.0, 0.0],  # D
            [0.0, 0.0],  # A
        ]
    )
    polyline_12345 = np.array(
        [
            [-2.0, -9.0, 0.0],  # 1
            [3.0, 7.0, 1.0],  # 2
            [3.0, -7.0, 2.0],  # 3
            [8.0, 7.0, 3.0],  # 4
            [8.0, -7.0, 4.0],  # 5
        ]
    )
    ruler = PolylineRuler(polyline_12345, is_wgs84=False)
    chunks = polyline_in_polygon(polyline_12345, polygon_ABCD)
    ranges = []
    for _, _, r1, _, _, r2 in chunks:
        ranges.append(r2 - r1)
    expected_ranges = [2.72883, 14.4676666, 7.01783]
    np.testing.assert_allclose(ranges, expected_ranges, atol=1e-4)

    # test inside
    polyline = next(iter(chunks.values()))
    polyline_updated = np.copy(polyline_12345)
    polyline_updated[0] = polyline.mean(axis=0)
    chunks2 = polyline_in_polygon(polyline_updated, polygon_ABCD)
    __print_chunks(chunks)
    __print_chunks(chunks2)

    chunks1 = list(chunks.items())
    chunks2 = list(chunks2.items())
    assert len(chunks1) == len(chunks2)
    assert np.fabs(chunks1[0][1][-1] - chunks2[0][1][-1]).max() == 0.0
    assert np.fabs(chunks1[1][1] - chunks2[1][1]).sum() < 1e-6
    assert np.fabs(chunks1[2][1] - chunks2[2][1]).sum() < 1e-6

    chunks3 = polyline_in_polygon(ruler.lineSliceAlong(1, 2), polygon_ABCD)
    assert len(chunks3) == 0
    chunks3 = polyline_in_polygon(ruler.lineSliceAlong(7, 8), polygon_ABCD)
    assert len(chunks3) == 1

    # test fc
    fc = FastCrossing()
    fc.add_polyline(polygon_ABCD)
    chunks2 = polyline_in_polygon(polyline_12345, polygon_ABCD, fc=fc)
    assert list(chunks.keys()) == list(chunks2.keys())
    N = 1000
    tick = time.time()
    for _ in range(N):
        polyline_in_polygon(polyline_12345, polygon_ABCD)
    delta1 = time.time() - tick
    tick = time.time()
    for _ in range(N):
        polyline_in_polygon(polyline_12345, polygon_ABCD, fc=fc)
    delta2 = time.time() - tick
    print(delta1, delta2)
    # assert delta2 < delta1

    anchor_lla = [123.4, 56.7, 8.9]
    chunks = polyline_in_polygon(
        tf.enu2lla(polyline_12345, anchor_lla=anchor_lla),
        tf.enu2lla(
            np.c_[polygon_ABCD, np.zeros(len(polygon_ABCD))], anchor_lla=anchor_lla
        )[:, :2],
        is_wgs84=True,
    )
    ranges = []
    for _, _, r1, _, _, r2 in chunks:
        ranges.append(r2 - r1)
    np.testing.assert_allclose(ranges, expected_ranges, atol=1e-4)

    chunks = polyline_in_polygon([[100, 0, 0], [200, 0, 0]], polygon_ABCD)
    assert len(chunks) == 0

    # TODO, test touches


def pytest_main(dir: str, *, test_file: str = None):
    os.chdir(dir)
    sys.exit(
        pytest.main(
            [
                dir,
                *(["-k", test_file] if test_file else []),
                "--capture",
                "tee-sys",
                "-vv",
                "-x",
            ]
        )
    )


def test_duplicates():
    fc = FastCrossing()
    fc.add_polyline(np.array([[0.0, 0.0], [5.0, 0.0]]))
    assert len(fc.intersections()) == 0
    fc.add_polyline(np.array([[0.0, 0.0], [5.0, 0.0]]))
    assert len(fc.intersections()) == 1


def test_flatbush():
    polyline = [
        [0, 0],
        [10, 0],
        [10, 5],
        [20, 5],
    ]
    fb = FlatBush()
    fb.add(polyline)
    fb.add(polyline[::-1])
    assert np.all(
        fb.labels()
        == [
            [-1, 0],
            [-1, 1],
            [-1, 2],
            [-1, 0],
            [-1, 1],
            [-1, 2],
        ]
    )
    assert np.all(
        fb.boxes()
        == [
            [0.0, 0.0, 10.0, 0.0],
            [10.0, 0.0, 10.0, 5.0],
            [10.0, 5.0, 20.0, 5.0],
            [20.0, 5.0, 10.0, 5.0],
            [10.0, 5.0, 10.0, 0.0],
            [10.0, 0.0, 0.0, 0.0],
        ]
    )

    assert not fb.search([0, 0], [10, 10])
    fb.finish()
    assert fb.search([0, 0], [10, 10])

    assert sorted(
        fb.search(
            [
                0,
                0,
            ],
            [0, 0],
        )
    ) == [0, 5]
    assert np.all(fb.box(0)[:2] == fb.box(5)[-2:])
    assert np.all(fb.box(0)[-2:] == fb.box(5)[:2])

    fb = FlatBush()
    fb.add(polyline)
    fb.add(polyline[::-1])
    fb.add(polyline)
    fb.add(polyline[::-1])
    fb.finish()
    assert sorted(
        fb.search(
            [
                0,
                0,
            ],
            [0, 0],
        )
    ) == [0, 5, 6, 11]


if __name__ == "__main__":
    np.set_printoptions(suppress=True)
    pwd = os.path.abspath(os.path.dirname(__file__))
    pytest_main(pwd, test_file=os.path.basename(__file__))



from __future__ import annotations

import numpy as np

from ._core import KdTree as _KdTree


class KDTree:
    def __init__(self, data: np.ndarray, leafsize: int = 10, *args, **kwargs):
        data = np.asarray(data, dtype=np.float64)
        self.tree: _KdTree = _KdTree(data)
        self.tree.set_leafsize(leafsize)

    @staticmethod
    def vec3(arr: np.ndarray):
        return np.r_[arr, 0.0] if len(arr) == 2 else np.asarray(arr, dtype=np.float64)

    def count_neighbors(self, *args, **kwargs):
        raise NotImplementedError

    def query(self, x, k=1, *args, **kwargs):
        x = np.asarray(x, dtype=np.float64)
        if x.ndim == 1:
            xyz = self.vec3(x)
            ii, dd = self.tree.nearest(xyz, k=k)
            return dd, ii
        if isinstance(k, (int, np.integer)):
            ret_ii, ret_dd = [], []
            for xyz in x:
                xyz = self.vec3(xyz)  # noqa: PLW2901
                if k == 1:
                    ii, dd = self.tree.nearest(xyz)
                else:
                    ii, dd = self.tree.nearest(xyz, k=k)
                    ii = ii.tolist()
                    dd = dd.tolist()
                ret_ii.append(ii)
                ret_dd.append(dd)
            return ret_dd, ret_ii
        K = max(k)
        ret_ii, ret_dd = [], []
        for xyz in x:
            xyz = self.vec3(xyz)  # noqa: PLW2901
            ii, dd = self.tree.nearest(xyz, k=K)
            ii = [ii[kk - 1] for kk in k]
            dd = [dd[kk - 1] for kk in k]
            ret_ii.append(ii)
            ret_dd.append(dd)
        return ret_dd, ret_ii

    def query_ball_point(
        self,
        x,
        r,
        p=2.0,
        eps=0,
        workers=1,
        return_sorted=None,
        return_length=False,
        *args,
        **kwargs,
    ):
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query_ball_point.html#scipy.spatial.cKDTree.query_ball_point
        """
        x = np.asarray(x, dtype=np.float64)
        if return_sorted is None:
            # If None, does not sort single point queries, but does sort
            # multi-point queries which was the behavior before this option was
            # added.
            return_sorted = x.ndim != 1
        if x.ndim == 1:
            xyz = self.vec3(x)
            ii, _ = self.tree.nearest(
                xyz,
                radius=r,
                sort=return_sorted,
                return_squared_l2=True,
            )
            if return_length:
                return len(ii)
            return ii.tolist()
        if return_sorted is None:
            return_sorted = True
        if isinstance(r, (int, float, np.number)):
            r = [r] * len(x)
        ret_ii = []
        for pp, rr in zip(x, r):
            xyz = self.vec3(pp)
            ii, _ = self.tree.nearest(
                xyz,
                radius=rr,
                sort=return_sorted,
                return_squared_l2=True,
            )
            ret_ii.append(ii.tolist())
        if return_length:
            ret_ii = [len(ii) for ii in ret_ii]
        return ret_ii

    def query_ball_tree(self, *args, **kwargs):
        raise NotImplementedError

    def query_pairs(self, *args, **kwargs):
        raise NotImplementedError

    def query_distance_matrix(self, *args, **kwargs):
        raise NotImplementedError


# create alias
cKDTree = KDTree
